@@ -292,10 +292,18 @@ check_message_id (struct mlfi_priv *priv, char *header)
 		upstream_ok (&selected->up, priv->conn_tm.tv_sec);
 		priv->strict = 0;
 		rmilter_strlcpy (priv->reply_id, header, sizeof (priv->reply_id));
+		msg_info ("check_message_id: %s: from %s[%s] from=<%s> to=<%s> is reply to our message %s; "
+				"skip dcc and spamd checks",
+						priv->mlfi_id,
+						priv->priv_hostname,
+						priv->priv_ip,
+						priv->priv_from,
+						priv->rcpts.lh_first->r_addr,
+						priv->reply_id);
 		return;
 	}
 	else if (r != NOT_EXISTS) {
-		msg_info ("mlfi_data: cannot read data from memcached: %s", memc_strerror (r));
+		msg_err ("mlfi_data: cannot read data from memcached: %s", memc_strerror (r));
 		upstream_fail (&selected->up, priv->conn_tm.tv_sec);
 		memc_close_ctx (&mctx);
 		return;
@@ -314,7 +322,7 @@ check_greylisting_ctx(SMFICTX *ctx, struct mlfi_priv *priv)
 	if (priv->priv_ip[0] != '\0' && cfg->memcached_servers_grey_num > 0 &&
 			cfg->greylisting_timeout > 0 && cfg->greylisting_expire > 0 && priv->strict != 0) {
 
-		msg_debug ("mlfi_data: %s: checking greylisting", priv->mlfi_id);
+		msg_debug ("check_greylisting_ctx: %s: checking greylisting", priv->mlfi_id);
 		ptr = priv->priv_addr.family == AF_INET6 ? (void *)&priv->priv_addr.addr.sa6.sin6_addr :
 				(void *)&priv->priv_addr.addr.sa4.sin_addr;
 		r = check_greylisting (cfg, ptr, priv->priv_addr.family, &priv->conn_tm,
@@ -322,14 +330,14 @@ check_greylisting_ctx(SMFICTX *ctx, struct mlfi_priv *priv)
 		switch (r) {
 		case GREY_GREYLISTED:
 			if (smfi_setreply (ctx, RCODE_LATER, XCODE_TEMPFAIL, cfg->greylisted_message) != MI_SUCCESS) {
-				msg_err("mlfi_data: %s: smfi_setreply failed", priv->mlfi_id);
+				msg_err("check_greylisting_ctx: %s: smfi_setreply failed", priv->mlfi_id);
 			}
 			CFG_UNLOCK();
 			return SMFIS_TEMPFAIL;
 			break;
 		case GREY_ERROR:
 			if (smfi_setreply (ctx, RCODE_TEMPFAIL, XCODE_TEMPFAIL, (char *)"Service unavailable") != MI_SUCCESS) {
-				msg_err("mlfi_data: %s: smfi_setreply failed", priv->mlfi_id);
+				msg_err("check_greylisting_ctx: %s: smfi_setreply failed", priv->mlfi_id);
 			}
 			CFG_UNLOCK();
 			return SMFIS_TEMPFAIL;
@@ -411,7 +419,7 @@ send_beanstalk_copy (const struct mlfi_priv *priv, struct beanstalk_server *srv)
 		return;
 	}
 	else {
-		msg_info ("send_beanstalk_copy: cannot put data to beanstalk: %s", bean_strerror (r));
+		msg_err ("send_beanstalk_copy: cannot put data to beanstalk: %s", bean_strerror (r));
 		upstream_fail (&srv->up, priv->conn_tm.tv_sec);
 		bean_close_ctx (&bctx);
 		return;
@@ -454,7 +462,8 @@ send_beanstalk (const struct mlfi_priv *priv)
 	}
 
 	if ((map = mmap (NULL, priv->eoh_pos, PROT_READ, MAP_SHARED, fd, 0)) == MAP_FAILED) {
-		msg_err ("send_beanstalk: cannot mmap file %s, %s", priv->file, strerror (errno));
+		msg_err ("send_beanstalk: %s: cannot mmap file %s: %s", priv->mlfi_id,
+				priv->file, strerror (errno));
 		close (fd);
 		return;
 	}
@@ -468,7 +477,8 @@ send_beanstalk (const struct mlfi_priv *priv)
 
 	r = bean_init_ctx (&bctx);
 	if (r == -1) {
-		msg_warn ("send_beanstalk: cannot connect to beanstalk upstream: %s",
+		msg_warn ("send_beanstalk: %s: cannot connect to beanstalk upstream: %s",
+				priv->mlfi_id,
 				inet_ntop (AF_INET, &selected->addr, ipout, sizeof (ipout)));
 		upstream_fail (&selected->up, priv->conn_tm.tv_sec);
 		munmap (map, priv->eoh_pos);
@@ -490,7 +500,7 @@ send_beanstalk (const struct mlfi_priv *priv)
 		return;
 	}
 	else {
-		msg_info ("send_beanstalk: cannot put data to beanstalk: %s", bean_strerror (r));
+		msg_err ("send_beanstalk: cannot put data to beanstalk: %s", bean_strerror (r));
 		upstream_fail (&selected->up, priv->conn_tm.tv_sec);
 		bean_close_ctx (&bctx);
 		return;
@@ -637,18 +647,19 @@ try_wildcard_dkim (const char *domain, struct mlfi_priv *priv)
 										(u_char *)keymap,  (u_char *)dkim_domain->selector,
 										(u_char *)domain,
 										cfg->dkim_relaxed_header ? DKIM_CANON_RELAXED : DKIM_CANON_SIMPLE,
-												cfg->dkim_relaxed_body ? DKIM_CANON_RELAXED : DKIM_CANON_SIMPLE,
-														cfg->dkim_sign_sha256 ? DKIM_SIGN_RSASHA256 : DKIM_SIGN_RSASHA1, -1, &statp);
+										cfg->dkim_relaxed_body ? DKIM_CANON_RELAXED : DKIM_CANON_SIMPLE,
+										cfg->dkim_sign_sha256 ? DKIM_SIGN_RSASHA256 : DKIM_SIGN_RSASHA1, -1, &statp);
 								/* It is safe to unmap memory here */
 								munmap (keymap, st.st_size);
 								if (statp != DKIM_STAT_OK) {
-									msg_info ("dkim sign failed: %d", statp);
+									msg_info ("dkim sign failed: %s", dkim_geterror (d));
 									if (d) {
 										dkim_free (d);
 									}
 									return NULL;
 								}
 								else {
+									priv->dkim_domain = dkim_domain;
 									return d;
 								}
 							}
@@ -693,6 +704,24 @@ mlfi_envfrom(SMFICTX *ctx, char **envfrom)
 	priv->priv_from[i] = '\0';
 	msg_debug ("mlfi_envfrom: got from value: %s", priv->priv_from);
 
+	if (priv->priv_hostname[0] == '\0') {
+		tmpfrom = smfi_getsymval(ctx, "{client_name}");
+		if (tmpfrom != NULL) {
+			rmilter_strlcpy (priv->priv_hostname, tmpfrom, sizeof (priv->priv_hostname));
+			msg_debug ("mlfi_envfrom: got host value: %s", priv->priv_hostname);
+		}
+		else {
+			rmilter_strlcpy (priv->priv_hostname, "unknown", sizeof (priv->priv_hostname));
+		}
+	}
+
+
+	tmpfrom = smfi_getsymval(ctx, "{auth_authen}");
+	if (tmpfrom != NULL) {
+		priv->authenticated = 1;
+		rmilter_strlcpy (priv->priv_user, tmpfrom, sizeof (priv->priv_user));
+	}
+
 	/* Check whether we need to sign this message */
 #ifdef WITH_DKIM
 	CFG_RLOCK();
@@ -709,7 +738,7 @@ mlfi_envfrom(SMFICTX *ctx, char **envfrom)
 		else {
 			HASH_FIND_STR (cfg->dkim_domains, domain_pos + 1, dkim_domain, strncasecmp);
 		}
-		if (!cfg->dkim_auth_only || *priv->priv_user != '\0') {
+		if (!cfg->dkim_auth_only || priv->authenticated) {
 			if (dkim_domain && dkim_domain->is_loaded) {
 				priv->dkim = dkim_sign (cfg->dkim_lib,  (u_char *)"rmilter", NULL,
 						(u_char *)dkim_domain->key,  (u_char *)dkim_domain->selector,
@@ -718,55 +747,32 @@ mlfi_envfrom(SMFICTX *ctx, char **envfrom)
 								cfg->dkim_relaxed_body ? DKIM_CANON_RELAXED : DKIM_CANON_SIMPLE,
 										cfg->dkim_sign_sha256 ? DKIM_SIGN_RSASHA256 : DKIM_SIGN_RSASHA1, -1, &statp);
 				if (statp != DKIM_STAT_OK) {
-					msg_info ("dkim sign failed: %d", statp);
+					msg_info ("dkim sign failed: %s", dkim_geterror (priv->dkim));
 					if (priv->dkim) {
 						dkim_free (priv->dkim);
 					}
 					priv->dkim = NULL;
 				}
 				else {
-					msg_info ("try to add signature for %s domain", dkim_domain->domain);
+					msg_debug ("try to add signature for %s domain", dkim_domain->domain);
+					priv->dkim_domain = dkim_domain;
 				}
 			}
 			else {
 				priv->dkim = try_wildcard_dkim (domain_pos + 1, priv);
 				if (priv->dkim) {
-					msg_info ("try to add signature for %s domain", domain_pos + 1);
+					msg_debug ("try to add signature for %s domain", domain_pos + 1);
 				}
 			}
 		}
 		else {
 			priv->dkim = NULL;
-			msg_info ("do not add dkim signature for unauthorized user");
+			msg_debug ("do not add dkim signature for unauthorized user");
 		}
 	}
 	CFG_UNLOCK();
 #endif
 
-	if (priv->priv_hostname[0] == '\0') {
-		tmpfrom = smfi_getsymval(ctx, "{client_name}");
-		if (tmpfrom != NULL) {
-			rmilter_strlcpy (priv->priv_hostname, tmpfrom, sizeof (priv->priv_hostname));
-			msg_debug ("mlfi_envfrom: got host value: %s", priv->priv_hostname);
-		}
-		else {
-			rmilter_strlcpy (priv->priv_hostname, "unknown", sizeof (priv->priv_hostname));
-		}
-	}
-
-
-	tmpfrom = smfi_getsymval(ctx, "{auth_authen}");
-	if (tmpfrom != NULL) {
-#ifndef STRICT_AUTH
-		if (!cfg->strict_auth) {
-			msg_info ("mlfi_envfrom: turn off strict checks for authenticated sender: %s", tmpfrom);
-			priv->strict = 0;
-		}
-#endif
-		rmilter_strlcpy (priv->priv_user, tmpfrom, sizeof (priv->priv_user));
-	}
-
-
 	CFG_RLOCK();
 	/* Check connect */
 	act = regexp_check (cfg, priv, STAGE_CONNECT);
@@ -868,12 +874,19 @@ mlfi_data(SMFICTX *ctx)
 	}
 	else {
 		rmilter_strlcpy (priv->mlfi_id, "NOQUEUE", sizeof (priv->mlfi_id));
-		msg_info ("mlfi_data: cannot get queue id, set to 'NOQUEUE'");
+		msg_err ("mlfi_data: cannot get queue id, set to 'NOQUEUE'");
 	}
 	CFG_UNLOCK();
 
+	if (priv->authenticated && !cfg->strict_auth) {
+		msg_info ("mlfi_envfrom: %s: turn off strict checks for authenticated sender: %s",
+				priv->mlfi_id, priv->priv_user);
+		priv->strict = 0;
+	}
+
 	if (!cfg->spamd_greylist) {
-		if ((r = check_greylisting_ctx (ctx, priv)) != SMFIS_CONTINUE) {
+		if (!priv->authenticated &&
+				(r = check_greylisting_ctx (ctx, priv)) != SMFIS_CONTINUE) {
 			msg_info ("mlfi_eom: %s: greylisting message", priv->mlfi_id);
 			mlfi_cleanup (ctx, false);
 			return r;
@@ -950,7 +963,7 @@ mlfi_header(SMFICTX * ctx, char *headerf, char *headerv)
 				snprintf ((char *)tmp, tmplen, "%s: %s", headerf, headerv);
 				r = dkim_header (priv->dkim, tmp, tmplen - 1);
 				if (r != DKIM_STAT_OK) {
-					msg_info ("dkim_header failed: %d", r);
+					msg_info ("dkim_header failed: %s", dkim_geterror (priv->dkim));
 				}
 				free (tmp);
 			}
@@ -995,7 +1008,7 @@ mlfi_eoh(SMFICTX * ctx)
 
 	if (!priv->fileh) {
 		if (create_temp_file (priv) == -1) {
-			msg_err ("mlfi_eoh: cannot create temp file");
+			msg_err ("mlfi_eoh: %s: cannot create temp file", priv->mlfi_id);
 			mlfi_cleanup (ctx, false);
 			return SMFIS_TEMPFAIL;
 		}
@@ -1017,7 +1030,7 @@ mlfi_eoh(SMFICTX * ctx)
 	if (priv->dkim) {
 		r = dkim_eoh (priv->dkim);
 		if (r != DKIM_STAT_OK) {
-			msg_info ("dkim_eoh failed: %d", r);
+			msg_info ("mlfi_eoh: %s: dkim_eoh failed: %s", priv->mlfi_id, dkim_geterror (priv->dkim));
 		}
 	}
 #endif
@@ -1058,7 +1071,7 @@ mlfi_eom(SMFICTX * ctx)
 	}
 	else {
 		rmilter_strlcpy (priv->mlfi_id, "NOQUEUE", sizeof (priv->mlfi_id));
-		msg_info ("mlfi_eom: cannot get queue id, set to 'NOQUEUE'");
+		msg_err ("mlfi_eom: cannot get queue id, set to 'NOQUEUE'");
 	}
 
 
@@ -1128,18 +1141,20 @@ mlfi_eom(SMFICTX * ctx)
 	}
 	else if (cfg->sizelimit != 0 && sb.st_size > (off_t)cfg->sizelimit) {
 #ifndef FREEBSD_LEGACY
-		msg_warn ("mlfi_eom: %s: message size(%zd) exceeds limit(%zd), not scanned, %s", priv->mlfi_id, (size_t)sb.st_size, cfg->sizelimit, priv->file);
+		msg_warn ("mlfi_eom: %s: message size(%zd) exceeds limit(%zd), not scanned, %s",
+				priv->mlfi_id, (size_t)sb.st_size, cfg->sizelimit, priv->file);
 #else
-		msg_warn ("mlfi_eom: %s: message size(%ld) exceeds limit(%ld), not scanned, %s", priv->mlfi_id, (long int)sb.st_size, (long int)cfg->sizelimit, priv->file);
+		msg_warn ("mlfi_eom: %s: message size(%ld) exceeds limit(%ld), not scanned, %s",
+				priv->mlfi_id, (long int)sb.st_size, (long int)cfg->sizelimit, priv->file);
 #endif
 		CFG_UNLOCK();
 		return mlfi_cleanup (ctx, true);
 	}
-	msg_info ("mlfi_eom: %s: tempfile=%s, size=%lu", priv->mlfi_id, priv->file, (unsigned long int)sb.st_size);
+	msg_info ("mlfi_eom: %s: tempfile=%s, size=%lu",
+			priv->mlfi_id, priv->file, (unsigned long int)sb.st_size);
 
 	if (!priv->strict) {
-		msg_info ("mlfi_eom: %s: from %s[%s] from=<%s> to=<%s> is reply to our message %s; skip dcc, spamd", priv->mlfi_id, 
-				priv->priv_hostname, priv->priv_ip, priv->priv_from, priv->rcpts.lh_first->r_addr, priv->reply_id);
+
 	}
 
 #ifdef HAVE_DCC
@@ -1171,48 +1186,13 @@ mlfi_eom(SMFICTX * ctx)
 	}
 #endif
 
-	/* Check clamav */
-	if (cfg->clamav_servers_num != 0) {
-		msg_debug ("mlfi_eom: %s: check clamav", priv->mlfi_id);
-		r = check_clamscan (priv->file, strres, sizeof (strres));
-		if (r < 0) {
-			msg_warn ("mlfi_eom: %s: check_clamscan() failed, %d", priv->mlfi_id, r);
-			CFG_UNLOCK();
-			mlfi_cleanup (ctx, false);
-			return SMFIS_TEMPFAIL;
-		}
-		if (*strres) {
-			msg_warn ("mlfi_eom: %s: rejecting virus %s", priv->mlfi_id, strres);
-			snprintf (buf, sizeof (buf), "Infected: %s", strres);
-			smfi_setreply (ctx, RCODE_REJECT, XCODE_REJECT, buf);
-			CFG_UNLOCK();
-			mlfi_cleanup (ctx, false);
-			return SMFIS_REJECT;
-		}
-	}
-	/* Write message to beanstalk */
-	if (cfg->beanstalk_servers_num > 0 && cfg->send_beanstalk_headers) {
-		send_beanstalk (priv);
-	}
-	/* Maybe write its copy */
-	if (cfg->copy_server && cfg->send_beanstalk_copy) {
-		prob_cur = cfg->beanstalk_copy_prob;
-		/* Normalize */
-		prob_max = 100;
-		while (prob_cur < 1.0) {
-			prob_max *= 10;
-			prob_cur *= 10;
-		}
-		if (rand () % prob_max <= prob_cur) {
-			send_beanstalk_copy (priv, cfg->copy_server);
-		}
-	}
 	if (priv->priv_addr.family == AF_INET) {
 		if (radix32tree_find (cfg->spamd_whitelist,
 				ntohl((uint32_t)priv->priv_addr.addr.sa4.sin_addr.s_addr)) != RADIX_NO_VALUE) {
 			ip_whitelisted = true;
 		}
 	}
+
 	/* Check spamd */
 	if (cfg->spamd_servers_num != 0 && !priv->has_whitelisted && priv->strict
 			&& !ip_whitelisted &&
@@ -1255,7 +1235,7 @@ mlfi_eom(SMFICTX * ctx)
 				if (r >= METRIC_ACTION_GREYLIST && cfg->spamd_greylist) {
 					/* Perform greylisting */
 					CFG_UNLOCK();
-					if (check_greylisting_ctx (ctx, priv) != SMFIS_CONTINUE) {
+					if (!priv->authenticated && check_greylisting_ctx (ctx, priv) != SMFIS_CONTINUE) {
 						msg_info ("mlfi_eom: %s: greylisting message according to spamd action", priv->mlfi_id);
 						mlfi_cleanup (ctx, false);
 						return SMFIS_TEMPFAIL;
@@ -1286,7 +1266,44 @@ mlfi_eom(SMFICTX * ctx)
 				}
 			}
 		}
+	}
 
+	/* Write message to beanstalk */
+	if (cfg->beanstalk_servers_num > 0 && cfg->send_beanstalk_headers) {
+		send_beanstalk (priv);
+	}
+	/* Maybe write its copy */
+	if (cfg->copy_server && cfg->send_beanstalk_copy) {
+		prob_cur = cfg->beanstalk_copy_prob;
+		/* Normalize */
+		prob_max = 100;
+		while (prob_cur < 1.0) {
+			prob_max *= 10;
+			prob_cur *= 10;
+		}
+		if (rand () % prob_max <= prob_cur) {
+			send_beanstalk_copy (priv, cfg->copy_server);
+		}
+	}
+
+	/* Check clamav */
+	if (cfg->clamav_servers_num != 0) {
+		msg_debug ("mlfi_eom: %s: check clamav", priv->mlfi_id);
+		r = check_clamscan (priv->file, strres, sizeof (strres));
+		if (r < 0) {
+			msg_warn ("mlfi_eom: %s: check_clamscan() failed, %d", priv->mlfi_id, r);
+			CFG_UNLOCK();
+			mlfi_cleanup (ctx, false);
+			return SMFIS_TEMPFAIL;
+		}
+		if (*strres) {
+			msg_warn ("mlfi_eom: %s: rejecting virus %s", priv->mlfi_id, strres);
+			snprintf (buf, sizeof (buf), "Infected: %s", strres);
+			smfi_setreply (ctx, RCODE_REJECT, XCODE_REJECT, buf);
+			CFG_UNLOCK();
+			mlfi_cleanup (ctx, false);
+			return SMFIS_REJECT;
+		}
 	}
 
 	/* Update rate limits for message */
@@ -1317,20 +1334,36 @@ mlfi_eom(SMFICTX * ctx)
 	char *hdr;
 	size_t len;
 	if (priv->dkim) {
-		dkim_set_margin (priv->dkim, 0);
+
+		if (!cfg->dkim_fold_header) {
+			/* Disable header folding */
+			dkim_set_margin (priv->dkim, 0);
+		}
+
 		r = dkim_eom (priv->dkim, NULL);
 		if (r == DKIM_STAT_OK) {
 			r = dkim_getsighdr_d (priv->dkim, 0, (u_char **)&hdr, &len);
 			if (r == DKIM_STAT_OK) {
-				msg_info ("<%s> added DKIM signature", priv->mlfi_id);
+				msg_info ("mlfi_eom: %s: d=%s, s=%s, added DKIM signature",
+						priv->mlfi_id,
+						dkim_getdomain (priv->dkim),
+						priv->dkim_domain->selector);
 				smfi_addheader (ctx, DKIM_SIGNHEADER, dkim_stripcr (hdr));
 			}
 			else {
-				msg_info ("<%s> sign failed: %d", priv->mlfi_id, r);
+				msg_info ("mlfi_eom: %s: d=%s, s=%s, sign failed: %s",
+						priv->mlfi_id,
+						dkim_getdomain (priv->dkim),
+						priv->dkim_domain->selector,
+						dkim_geterror (priv->dkim));
 			}
 		}
 		else {
-			msg_info ("<%s> dkim_eom failed: %d", priv->mlfi_id, r);
+			msg_info ("mlfi_eom: %s: d=%s, s=%s, dkim_eom failed: %s",
+					priv->mlfi_id,
+					dkim_getdomain (priv->dkim),
+					priv->dkim_domain->selector,
+					dkim_geterror (priv->dkim));
 		}
 	}
 #endif
@@ -1431,7 +1464,7 @@ mlfi_body(SMFICTX * ctx, u_char * bodyp, size_t bodylen)
 
 	if (!priv->fileh) {
 		if (create_temp_file (priv) == -1) {
-			msg_err ("mlfi_eoh: cannot create temp file");
+			msg_err ("mlfi_body: %s: cannot create temp file", priv->mlfi_id);
 			mlfi_cleanup (ctx, false);
 			return SMFIS_TEMPFAIL;
 		}
@@ -1459,7 +1492,8 @@ mlfi_body(SMFICTX * ctx, u_char * bodyp, size_t bodylen)
 	if (priv->dkim) {
 		r = dkim_body (priv->dkim, bodyp, bodylen);
 		if (r != DKIM_STAT_OK) {
-			msg_info ("<%s>: dkim_body failed: %d", priv->mlfi_id, r);
+			msg_info ("mlfi_body: %s: dkim_body failed: %s", priv->mlfi_id,
+					dkim_geterror (priv->dkim));
 		}
 	}
 #endif
